package gdefalll.gui;

import java.awt.Color;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Iterator;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.border.LineBorder;
import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import gdefalll.Grammars.PlanGrammarLexer;
import gdefalll.Grammars.PlanGrammarParser;
import gdefalll.events.PlanChangedGenerator;
import gdefalll.events.StringContentListener;
import gdefalll.gui.actions.*;
import gdefalll.gui.dialogs.PlanNameEditor;

/**
 * Plan viewer panel. It contains sequence of actions.
 * License: BUT OPEN SOURCE LICENCE v.1
 * 
 * @author Jan Horacek <ihoracek@fit.vutbr.cz>
 */
public class GuiPlan extends GuiDragable implements StringContentListener {

	private static final long serialVersionUID = 0L;

	private Color CPlan = new Color(245, 245, 255);
	private Color CPlanBorder = new Color(0, 0, 0);
	private Color CPlanDraggArea = new Color(140, 180, 255);
	private Color CIntentionDraggArea = new Color(250, 180, 255);
	private String _name = null;
	private int width = 250;
	private int height = 26;
	private Point _drag_start_point = new Point();
	private ArrayList<ALLLAction> actionList = new ArrayList<ALLLAction>();
	JPanel panel = new JPanel();
	JPanel title = new JPanel();
	JLabel title_label = new JLabel();
	/**
	 * Used to create and highlight empty space for new action.
	 */
	NewAction new_action = new NewAction();
	boolean selected = false;
	
	/**
	 * All listeners should register here to get an event when plan is changed.
	 */
	public PlanChangedGenerator planChanged = new PlanChangedGenerator();

	/**
	 * Constructor of the GUI Node objects.
	 * 
	 * @param node
	 */
	public GuiPlan(String name, int pos_x, int pos_y) {
		_name = name;
		setLayout(null);
		setX(pos_x);
		setY(pos_y);
		setSize(width, height);
		panel.setLayout(null);
		panel.setBackground(CPlan);
		panel.setBounds(0, 0, width, height);
		panel.setBorder(new LineBorder(CPlanBorder, 1, false));
		add(panel);
		// setOpaque(false);
		title.setLayout(null);
		if (name.equals("Intention"))
		{
			title.setBackground(CIntentionDraggArea);
		}
		else
		{
			title.setBackground(CPlanDraggArea);
		}		
		title.setBounds(2, 2, width - 4, 20);
		title.setBorder(new LineBorder(CPlanBorder, 1, false));
		panel.add(title);

		//addAction(new AddBB(0));
		//addAction(new DelBB(1));
		//addAction(new AddBB(2));
		title_label.setText(name);
		title_label.setBounds(2, 0, 142, 20);
		setOpaque(true);
		title.add(title_label);
		new_action.setVisible(false);
		panel.add(new_action);
	}
	
	/**
	 * @see setSelected(boolean selected, int x, int y)
	 */
	public boolean setSelected(boolean selected)
	{
		return setSelected(selected, 0, 0);
	}

	/**
	 * Create space for new action to be added into plan, shift action indexes only when such shift is necessary.
	 * @param selected If plan should be selected to add new action at x,y position.
	 * @param x Mouse position in plan.
	 * @param y Mouse position in plan.
	 * @return If something has changed and plan should be repainted.
	 */
	public boolean setSelected(boolean selected, int x, int y) {
		if (this.selected != selected || selected) {
			this.selected = selected;
			if (selected)
			{
				height = ALLLAction.drag_indent + (actionList.size() + 1) * (ALLLAction.height + ALLLAction.action_indent);
				if (new_action.isAtPos(x, y))
				{
					return false;
				}
				ALLLAction action = actionAtPos(x, y);
				if (action != null)
				{
					setNewActionIndex(action.getIndex() + action.getIndexIndent());
				}
				else
				{
					setNewActionIndex(-1);
				}
				new_action.setVisible(true);
			}
			else
			{
				resizeHeight();
			}
			setSize(width, height);
			panel.setBounds(0, 0, width, height);
			return true;
		}
		return false;
	}
	
	/**
	 * Compute height of panel. Height is based on count of actions.
	 */
	private void resizeHeight()
	{
		height = ALLLAction.drag_indent + actionList.size() * (ALLLAction.height + ALLLAction.action_indent);
		new_action.setVisible(false);
		new_action.setIndex(actionList.size());
		removeIndentOfActions();
	}
	
	/**
	 * @param If index of new_action changes, then shift proper actions to move empty space for new action.
	 */
	private void setNewActionIndex(int index)
	{
		if (new_action.getIndex() != index) {
			if (index < 0)
			{
				new_action.setIndex(actionList.size());
				removeIndentOfActions();
			}
			else
			{
				new_action.setIndex(index);
				Iterator<ALLLAction> ai = actionList.iterator();
				while (ai.hasNext()) {
					ALLLAction action = ai.next();
					action.setNewActionIndex(index);
				}
			}
		}
	}
	
	/**
	 * Removes shift of all actions in plan. Usually called when we go to VIEW mode.
	 */
	private void removeIndentOfActions()
	{
		Iterator<ALLLAction> ai = actionList.iterator();
		while (ai.hasNext()) {
			ALLLAction action = ai.next();
			action.setIndexIndent(0);
		}
	}
	

	/**
	 * Add new action to the plan. New action is placed to the empty space created with setSelected(boolean selected, int x, int y).
	 * @param action
	 * @see setSelected(boolean selected, int x, int y)
	 */
	synchronized public void addAction(ALLLAction action) {
		int new_index = new_action.getIndex();
		action.setIndex(new_index);
		Iterator<ALLLAction> ai = actionList.iterator();
		while (ai.hasNext()) {
			ALLLAction old_action = ai.next();
			int old_action_index = old_action.getIndex();
			if (old_action_index >= new_index)
			{
				old_action.setIndex(old_action_index + 1);
			}
		}
		actionList.add(action.getIndex(),action);
		new_action.setIndex(actionList.size());
		panel.add(action);
		setSelected(false);
		repaint();
	}
	
	/**
	 * Put new action to the plan at the position given by index.
	 * @param action
	 * @param index
	 */
	synchronized public void addAction(ALLLAction action, int index) {
		action.setIndex(index);
		Iterator<ALLLAction> ai = actionList.iterator();
		while (ai.hasNext()) {
			ALLLAction old_action = ai.next();
			int old_action_index = old_action.getIndex();
			if (old_action_index >= index)
			{
				old_action.setIndex(old_action_index + 1);
			}
		}
		actionList.add(action.getIndex(),action);
		new_action.setIndex(actionList.size());
		panel.add(action);
		setSelected(false);
		repaint();
	}
	
	synchronized public void removeAction(ALLLAction action) {
		if (action != null)
		{
			int action_index = action.getIndex();
			Iterator<ALLLAction> ai = actionList.iterator();
			while (ai.hasNext()) {
				ALLLAction old_action = ai.next();
				int old_action_index = old_action.getIndex();
				if (old_action_index >= action_index)
				{
					old_action.setIndex(old_action_index - 1);
				}
			}
			actionList.remove(action);
			panel.remove(action);
			resizeHeight();
			setSize(width, height);
			panel.setBounds(0, 0, width, height);
			repaint();
		}		
	}
	
	/**
	 * Removes actions from index indexFrom to index indexTo. It swaps them automatically when (indexFrom > indexTo).
	 * @param indexFrom
	 * @param indexTo
	 */
	synchronized public void removeAction(int indexFrom, int indexTo) {
		Iterator<ALLLAction> ai = actionList.iterator();
		ArrayList<ALLLAction> removedActions = new ArrayList<ALLLAction>();
		while (ai.hasNext()) {
			ALLLAction action = ai.next();
			int action_index = action.getIndex();
			if ((action_index >= indexFrom && action_index <= indexTo) || (action_index >= indexTo && action_index <= indexFrom))
			{
				removedActions.add(action); //Can't be removed directly here, otherwise it brakes iterator. We select actions to be removed.
			}
		}
		ai = removedActions.iterator();
		while (ai.hasNext()) {
			ALLLAction action = ai.next();
			int action_index = action.getIndex();
			if ((action_index >= indexFrom && action_index <= indexTo) || (action_index >= indexTo && action_index <= indexFrom))
			{
				removeAction(action);
			}
		}
	}

	/**
	 * Give action at position [pos_x, pos_y]. Usually called when we click with mouse on some action.
	 * @param pos_x
	 * @param pos_y
	 */
	public ALLLAction actionAtPos(int pos_x, int pos_y) {
		Iterator<ALLLAction> ai = actionList.iterator();
		while (ai.hasNext()) {
			ALLLAction action = ai.next();
			if (action.isAtPos(pos_x, pos_y)) {
				return action;
			}
		}
		return null;
	}

	/**
	 * Give true if position [pos_x, pos_y] is over a dragable area of the plan.
	 * Usually called when we click with mouse on title area (edit name/drag plan).		
	 * @param pos_x
	 * @param pos_y
	 */
	public boolean isAtDraggablePos(int pos_x, int pos_y) {
		return (pos_x > getX() + 2 && pos_x <= getX() + width - 4
				&& pos_y > getY() + 2 && pos_y <= getY() + 2 + 20);
	}

	/**
	 * @param pos_x
	 * @param pos_y
	 * @return True if some part of plan (title, actions) is at position [pos_x, pos_y].
	 */
	public boolean isAtPos(int pos_x, int pos_y) {
		return (pos_x > getX() && pos_x <= getX() + width && pos_y > getY() && pos_y <= getY()
				+ height);
	}

	/* (non-Javadoc)
	 * Gets name of plan.
	 * @see java.awt.Component#getName()
	 */
	public String getName() {
		return _name;
	}

	/**
	 * Set the position from where we started dragging the plan.
	 * @param start_x
	 * @param start_y
	 */
	public void setdragStartPoint(int start_x, int start_y) {
		_drag_start_point.x = start_x;
		_drag_start_point.y = start_y;
	}

	/**
	 * @return The position from where we started dragging the plan.
	 */
	public Point getdragStartPoint() {
		return _drag_start_point;
	}

	/* (non-Javadoc)
	 * When name of plan is modified from dialog.
	 * @see gdefalll.events.StringContentListener#stringContentReceived(java.lang.String)
	 */
	@Override
	public void stringContentReceived(String text) {
		_name = text;
		title_label.setText(text);
		this.planChanged.firePlanChanged();
	}

	/**
	 * Shows editor to edit name of plan.
	 */
	public void showEditor() {
		Point position = new Point(getLocation());
		SwingUtilities.convertPointToScreen(position, this);
		PlanNameEditor editor = new PlanNameEditor();
		position.x = position.x - editor.getWidth() / 2 + this.getWidth() / 2;
		position.y = position.y - editor.getHeight() / 2 + this.getHeight() / 2;
		editor.setLocation(position);
		editor.setText(title_label.getText());
		editor.sg.addListener(this);
		editor.setVisible(true);
	}
	
	/**
	 * @return ALLL code of this plan.
	 */
	public String getPlanCode()
	{
		Iterator<ALLLAction> ai = actionList.iterator();
		String output = "";
		while (ai.hasNext()) {
			ALLLAction action = ai.next();
			output += action.getALLLCode();
		}
		return output;
	}
	
	/**
	 * Set code of this plan. Called when we load some agent from file.
	 * @param code
	 * @param plan_editor Necesary for RemotePlan action to be set properly.
	 */
	public void setCode(String code, GuiPlanEditor plan_editor)
	{
		//System.out.println(code);
		ANTLRStringStream in = new ANTLRStringStream(code);
		PlanGrammarLexer lexer = new PlanGrammarLexer(in);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		//System.out.println("Lexer errors:" + lexer.getNumberOfSyntaxErrors());
		PlanGrammarParser parser = new PlanGrammarParser(tokens);
		try {
			parser.parse();
		} catch (RecognitionException e) {
			System.err.println("Problem at:" + e.charPositionInLine);
		}
		//System.out.println("Parser errors:" + parser.getNumberOfSyntaxErrors());
		ArrayList<String> errors = parser.getErrors();
		Iterator<String> ei = errors.iterator();
		while (ei.hasNext()) {
			String error = ei.next();
			System.err.println("Parser errors:" + error);
		}
		ArrayList<ALLLAction> actions = parser.getActionList();
		Iterator<ALLLAction> ai = actions.iterator();
		int index = 0;
		while (ai.hasNext()) {
			ALLLAction action = ai.next();
			if (action.getIdentifier().equals("^"))
			{
				((RemotePlan)action).setPlanList(plan_editor.getPlanList());
				plan_editor.planChanged.addListener((RemotePlan)action);
			}
			addAction(action, index);
			index++;
			//System.out.println("Adding action:" + action.getIdentifier());
		}
		resizeHeight();	
		setSize(width, height);
		panel.setBounds(0, 0, width, height);
		repaint();
		this.planChanged.firePlanChanged();
	}
}
