package gdefalll.gui;

import javax.swing.JPanel;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.JScrollPane;
import javax.swing.SwingUtilities;

import gdefalll.gui.GuiPlan;
import gdefalll.gui.actions.*;
import gdefalll.events.ActionChangedListener;
import gdefalll.events.PlanChangedGenerator;
import gdefalll.events.PlanChangedListener;
import gdefalll.events.StatusChangeGenerator;

/**
 * Plan editor component. It has some modes that add new actions and VIEW mode.
 * License: BUT OPEN SOURCE LICENCE v.1
 * 
 * @author Jan Horacek <ihoracek@fit.vutbr.cz>
 */
public class GuiPlanEditor extends JPanel implements MouseListener,
		MouseMotionListener, ComponentListener, ActionChangedListener,
		PlanChangedListener {

	/**
	 * Available modes of this component.
	 */
	public enum Mode {
		VIEW, ADD_BB, DEL_BB, CHECK_BB, CHECK_IB, WAIT_TIME, WAIT_MSG, WAIT_MSG_START, LEDS, MESSAGE, SUBPLAN, REMOTE_PLAN, CATCH_SECTION
	};

	/**
	 * Mode of component.
	 * @see enum Mode.
	 */
	private Mode mode = Mode.VIEW;

	private static final long serialVersionUID = 0L;

	/**
	 * Output for status bar.
	 */
	public StatusChangeGenerator status = new StatusChangeGenerator();

	/**
	 * List of plans that creates agent. First one is Intention.
	 */
	private ArrayList<GuiPlan> planList = new ArrayList<GuiPlan>();
	
	/**
	 * Which plan is dragged actually.
	 */
	private GuiPlan planDragged = null;
	
	private static final Dimension DefaultSizeX = new Dimension(4000, 3000);
	JScrollPane scrollPane = new JScrollPane(this);
	private Color lineCBright = new Color(250, 250, 250);
	private Color lineCDark = new Color(228, 228, 228);

	/**
	 * Event generator, listeners should register here.
	 */
	public PlanChangedGenerator planChanged = new PlanChangedGenerator();

	/**
	 * Create plan editor.
	 */
	public GuiPlanEditor() {
		setPreferredSize(DefaultSizeX);
		addMouseListener(this);
		addMouseMotionListener(this);
		setLayout(null);
		scrollPane.getHorizontalScrollBar().setValue(1); //non-sense line, but it has to be done, otherwise scrollbars are screwed.
		scrollPane.getVerticalScrollBar().setValue(1200);
		scrollPane.getHorizontalScrollBar().setValue(1400);
	}

	/**
	 * Set mode of plan editor.
	 * @param mode
	 * @see enum Mode.
	 */
	public void setMode(Mode mode) {
		this.mode = mode;
	}

	/**
	 * @return mode of plan editor.
	 * @see enum Mode.
	 */
	public Mode getMode() {
		return this.mode;
	}

	/* (non-Javadoc)
	 * Paint component and all plans.
	 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
	 */
	@Override
	public void paintComponent(Graphics g) {
		/* Parent paintComponent method */
		super.paintComponent(g);
		/* Draw Background */
		drawBackground(g);
		drawPlans(g);
	}

	/**
	 * Draw background grid.
	 * @param g
	 */
	private void drawBackground(Graphics g) {
		g.setColor(Color.white);
		g.fillRect(0, 0, getWidth(), getHeight());

		Dimension d = getSize();
		int width = (int) d.getWidth();
		int height = (int) d.getHeight();

		/* Vertical lines */
		for (int i = 0; i <= width; i += 10) {
			Color c = (i % 50 == 0) ? lineCDark : lineCBright;
			g.setColor(c);
			g.drawLine(i, 0, i, height);
		}
		/* Horizontal lines */
		for (int i = 0; i <= height; i += 10) {
			Color c = (i % 50 == 0) ? lineCDark : lineCBright;
			g.setColor(c);
			g.drawLine(0, i, width, i);
		}
	}

	/**
	 * Paint all plans of agent.
	 * @param g
	 */
	private void drawPlans(Graphics g) {
		Iterator<GuiPlan> pi = planList.iterator();
		while (pi.hasNext()) {
			pi.next().repaint();
			// pi.next().paintComponent(g);
		}
		// TODO prekryvan planu?
		// if (planDragged != null) {
		// planDragged.repaint();
		// }
	}

	/**
	 * @param x
	 * @param y
	 * @return Returns plan with title bar at position [x,y].
	 */
	private GuiPlan planAtDragablePos(int x, int y) {
		Iterator<GuiPlan> pi = planList.iterator();
		while (pi.hasNext()) {
			GuiPlan plan = pi.next();
			if (plan.isAtDraggablePos(x, y)) {
				return plan;
			}
		}
		return null;
	}

	/**
	 * @param x
	 * @param y
	 * @return Which plan is at [x,y] position.
	 */
	private GuiPlan planAtPos(int x, int y) {
		Iterator<GuiPlan> pi = planList.iterator();
		while (pi.hasNext()) {
			GuiPlan plan = pi.next();
			if (plan.isAtPos(x, y)) {
				return plan;
			}
		}
		return null;
	}

	/* (non-Javadoc)
	 * Double-click = edit action properties/plan name.
	 * Left-click = add new action / plan.
	 * Right-click = remove item.
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseClicked(MouseEvent e) {
		if (e.getClickCount() == 2) {
			GuiPlan plan = planAtPos(e.getX(), e.getY());
			if (plan != null) {
				if (plan.isAtDraggablePos(e.getX(), e.getY())) {
					plan.showEditor();
				} else {
					ALLLAction action = plan.actionAtPos(
							e.getX() - plan.getX(), e.getY() - plan.getY());
					if (action != null) {
						action.showEditor();
					}
				}
			}
		} else if (e.getClickCount() == 1) {
			if (SwingUtilities.isLeftMouseButton(e)) {
				GuiPlan plan = planAtPos(e.getX(), e.getY());
				if (plan != null) {
					if (this.mode != GuiPlanEditor.Mode.VIEW) // if we add some new action
					{
						ALLLAction action = null;
						switch (this.mode) {
							case ADD_BB:
								action = new AddBB();
								break;
							case DEL_BB:
								action = new DelBB();
								break;
							case CHECK_BB:
								action = new CheckBB();
								break;
							case CHECK_IB:
								action = new CheckInputBase();
								break;
							case WAIT_TIME:
								action = new WaitTime();
								break;
							case WAIT_MSG:
								action = new WaitMsg();
								break;
							case WAIT_MSG_START:
								action = new StartMsgWait();
								break;
							case LEDS:
								action = new Leds();
								break;
							case MESSAGE:
								action = new MessageOutput();
								break;
							case CATCH_SECTION:
								action = new CatchSection();
								break;
							case REMOTE_PLAN:
								action = new RemotePlan(this.planList);
								planChanged.addListener((RemotePlan)action);
								break;
							case SUBPLAN:
								InternalPlanEnd actionStop = new InternalPlanEnd();
								InternalPlanStart actionStart = new InternalPlanStart();
								actionStop.setNestedAction((ALLLAction)actionStart);
								actionStart.setNestedAction((ALLLAction)actionStop);
								addAction(plan, actionStop);
								plan.setSelected(true, e.getX() - plan.getX(), e.getY()	- plan.getY()); //necesarry to make space for second action
								action = actionStart;
								break;
						}
						addAction(plan, action);
					}
				} else {
					addEmptyPlan(e);
				}
			} else {
				GuiPlan plan = planAtDragablePos(e.getX(), e.getY());
				if (plan != null) {
					delPlan(planAtDragablePos(e.getX(), e.getY()));
				}
				else
				{
					GuiPlan planOfAction = planAtPos(e.getX(), e.getY());
					if (planOfAction != null)
					{
						removeAction(planOfAction, e.getX() - planOfAction.getX(), e.getY()	- planOfAction.getY());
					}					
				}				
			}
		}
		repaint();
	}
	
	/**
	 * Removes action from plan at position [x, y].
	 * @param plan
	 * @param x
	 * @param y
	 */
	synchronized private void removeAction(GuiPlan plan, int x, int y) {
		if (plan != null)
		{
			ALLLAction action = plan.actionAtPos(x, y);			
			if (action.getIdentifier().equals("@"))
			{
				ALLLAction nestedAction = ((InternalPlan) action).getNestedAction();
				plan.removeAction(action.getIndex(), nestedAction.getIndex());
			}
			else
			{
				plan.removeAction(action);
			}
			this.planChanged.firePlanChanged();
		}		
	}

	/**
	 * Add action to the plan. There has to be pre-made space in the plan with plan.setSelected().
	 * @param plan
	 * @param action
	 */
	private void addAction(GuiPlan plan, ALLLAction action) {
		if (action != null) {
			action.ACG.addListener(this);
			plan.addAction(action);
		}
		this.setMode(GuiPlanEditor.Mode.VIEW);
		this.planChanged.firePlanChanged();
	}

	@Override
	public void mouseEntered(MouseEvent e) {
		// Auto-generated method stub

	}

	@Override
	public void mouseExited(MouseEvent e) {
		// Auto-generated method stub
	}

	private int press_x, press_y;

	/*
	 * Start of Dragging. It sets planDragged value to dragged plan.
	 * @see mouseDragged().
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	@Override
	public void mousePressed(MouseEvent e) {
		press_x = e.getX();
		press_y = e.getY();
		planDragged = planAtDragablePos(e.getX(), e.getY());
		if (planDragged != null) {
			planDragged.setdragStartPoint(e.getX(), e.getY());
			remove(planDragged); // move window to the front
			add(planDragged, 0);
		}
		repaint();
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		if (planDragged == null) {

		}
		planDragged = null;
		repaint();
	}

	/*
	 * Moves plan that should be dragged. If mouse is outside any plan, it moves background.
	 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseDragged(MouseEvent e) {
		if (planDragged != null) {
			int dx = e.getX() - planDragged.getdragStartPoint().x;
			int dy = e.getY() - planDragged.getdragStartPoint().y;
			planDragged.setdragStartPoint(e.getX(), e.getY());
			planDragged.moveX(dx);
			planDragged.moveY(dy);
		} else {
			int dx = e.getX() - press_x;
			int dy = e.getY() - press_y;
			int scroll_x = getScrollPane().getHorizontalScrollBar().getValue();
			int scroll_y = getScrollPane().getVerticalScrollBar().getValue();
			getScrollPane().getHorizontalScrollBar().setValue(scroll_x - dx);
			getScrollPane().getVerticalScrollBar().setValue(scroll_y - dy);
		}
		repaint();
	}

	/*
	 * Highlights possition to add new action if we are in such mode.
	 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseMoved(MouseEvent e) {
		if (this.mode != Mode.VIEW) {
			int x = e.getX();
			int y = e.getY();
			Iterator<GuiPlan> pi = planList.iterator();
			while (pi.hasNext()) {
				GuiPlan plan = pi.next();
				if (plan.isAtPos(x, y)) {
					status.fire("Plan: " + plan.getName());
					if (plan.setSelected(true, e.getX() - plan.getX(), e.getY()
							- plan.getY())) {
						repaint();
					}
				} else {
					if (plan.setSelected(false)) {
						repaint();
					}
				}
			}
		}
	}

	// ===========================================================================================================================================================
	// Component Handler
	// ===========================================================================================================================================================

	@Override
	public void componentHidden(ComponentEvent e) {
		// Auto-generated method stub

	}

	@Override
	public void componentMoved(ComponentEvent e) {
		// Auto-generated method stub

	}

	@Override
	public void componentResized(ComponentEvent e) {
		// Auto-generated method stub

	}

	@Override
	public void componentShown(ComponentEvent e) {
		repaint();
	}

	// ===========================================================================================================================================================
	// Mouse Event Handler Functions
	// ===========================================================================================================================================================

	private int counter = 0;

	/**
	 * Add new plan at the mouse possition.
	 * @param e
	 */
	private void addEmptyPlan(MouseEvent e) {
		GuiPlan plan;
		if (counter > 0) {
			plan = new GuiPlan("plan" + counter, e.getX(), e.getY());
		} else {
			plan = new GuiPlan("Intention", e.getX(), e.getY());
		}
		plan.planChanged.addListener(this);
		counter++;
		planList.add(plan);
		add(plan);
		this.planChanged.firePlanChanged();
	}
	
	/**
	 * Add new plan. Plan position should be set in the plan properties!
	 * @param plan
	 */
	synchronized public void addPlan(GuiPlan plan) {		
		plan.planChanged.addListener(this);
		counter++;
		planList.add(plan);
		add(plan);
		this.planChanged.firePlanChanged();
	}

	/**
	 * Delete a plan.
	 * @param plan
	 */
	private void delPlan(GuiPlan plan) {
		if (plan != null) {
			if (plan != planList.get(0)) {
				planList.remove(plan);
				remove(plan);
			}
		}
		repaint();
		this.planChanged.firePlanChanged();
	}

	/**
	 * @return Scrolling area of this component.
	 */
	public JScrollPane getScrollPane() {
		return scrollPane;
	}

	/**
	 * @return Plan base of an agent.
	 */
	public String getPlanBase() {
		Iterator<GuiPlan> pi = planList.iterator();
		String output = "";
		if (pi.hasNext()) { //first action is Intention
			pi.next();
		}
		while (pi.hasNext()) {
			GuiPlan plan = pi.next();
			output += "(" + plan.getName() + ",(" + plan.getPlanCode() + "))";
		}
		return output;
	}

	/**
	 * @return Intention of an agent.
	 */
	public String getIntention() {
		Iterator<GuiPlan> pi = planList.iterator();
		String output = "";
		if (pi.hasNext()) {
			GuiPlan plan = pi.next();
			output += plan.getPlanCode();
		}
		return output;
	}

	/*
	 * Distribute event that agent changed to listeners.
	 * @see gdefalll.events.ActionChangedListener#actionChanged()
	 */
	@Override
	public void actionChanged() {
		this.planChanged.firePlanChanged();
	}

	/*
	 * Distribute event that agent changed to listeners.
	 * @see gdefalll.events.PlanChangedListener#planChanged()
	 */
	@Override
	public void planChanged() {
		this.planChanged.firePlanChanged();
	}
	
	/**
	 * @return Plan list.
	 */
	public ArrayList<GuiPlan> getPlanList()
	{
		return planList;
	}
}
