package gdefalll.gui.actions;

import gdefalll.events.ActionChangedGenerator;
import java.awt.Color;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;

/**
 * Elemental class of all ALLL actions.
 * Abstract class, each action type should implement abstract methods.
 * License: BUT OPEN SOURCE LICENCE v.1
 * 
 * @author Jan Horacek <ihoracek@fit.vutbr.cz>
 */
public abstract class ALLLAction extends JPanel{
	private static final long serialVersionUID = 0L;
	
	public static final int height = 20;
	private int width = 244;
	private static final int label_width = 120;
	public static int action_indent = 6;
	public static int drag_indent = 26;
	private static Color bg = new Color(255, 255, 255);
	private int index = -1;
	private int index_indent = 0;
	/**
	 * Label that shows action text identifier.
	 * @see getLabelText
	 */
	JLabel label;
	
	/**
	 * Generator of events that action has been changed. Listeners should register here.
	 */
	public ActionChangedGenerator ACG = new ActionChangedGenerator();
	
	/**
	 * Constructor.
	 * It creates label describing the action only.
	 * Value label should be defined inside each type of final action.
	 */
	public ALLLAction()
	{
		super();
		setLayout(null);
		setBackground(bg);
		setIndex(0);
		
		label = new JLabel(getLabelText());
		label.setForeground(getLabelColor());
		label.setIcon(getIcon());
		label.setBounds(0, 0, label_width, height);
		add(label);
		setOpaque(true);
	}
	
	/**
	 * Set position in plan.
	 * @param index Position in plan. Counter starts with 0.
	 */
	public void setIndex(int index)
	{
		if (this.index != index)
		{
			this.index = index;
			int pos_x = 3;
			int pos_y = drag_indent + (index + index_indent) * (height + action_indent);
			setBounds(pos_x, pos_y, width, height);
			repaint();
		}
	}
	
	/**
	 * Shift action index for desired indent. Used when new action place is highlighted.
	 * @param index_indent
	 */
	public void setIndexIndent(int index_indent)
	{
		if (this.index_indent != index_indent)
		{
			this.index_indent = index_indent;
			int pos_x = 3;
			int pos_y = drag_indent + (index + index_indent) * (height + action_indent);
			setBounds(pos_x, pos_y, width, height);
			repaint();
		}
	}
	
	/**
	 * Shift action by 1 when we are after new action empty space.
	 * @param index_of_new_action
	 * @see setIndexIndent
	 */
	public void setNewActionIndex(int index_of_new_action)
	{
		if (this.index < index_of_new_action)
		{
			setIndexIndent(0);
		}
		else
		{
			setIndexIndent(1);
		}
	}
	
	/**
	 * @return Position of action in the plan.
	 */
	public int getIndex()
	{
		return index;
	}
	
	/**
	 * @return Indent of action.
	 */
	public int getIndexIndent()
	{
		return index_indent;
	}
	
	/**
	 * @param x
	 * @param y
	 * @return True, if the action is at position [x, y].
	 */
	public boolean isAtPos(int x, int y)
	{
		if (x >= getX()-3 && x < getX()+4 + width && y >= getY()-3 && y < getY()+height+3)
		{
			return true;
		}
		else
		{
			return false;
		}		
	}
	
	
	
	/**
	 * Show editor to edit action properties.
	 */
	abstract public void showEditor();
	/**
	 * @return Get icon of this specific action type.
	 */
	abstract ImageIcon getIcon();
	/**
	 * @return Get identifier of this specific action type.
	 */
	abstract public String getIdentifier();
	/**
	 * @return Get label describing this specific action type.
	 */
	abstract String getLabelText();
	/**
	 * @return Get color of label of this specific action type.
	 */
	abstract Color getLabelColor();
	/**
	 * @return Get ALLL code of this specific action type.
	 */
	abstract public String getALLLCode();
	/**
	 * Set parameters of this specific action type.
	 * @param params
	 */
	abstract public void setParameters(String params);
}
