package gdefalll.gui.highlighters;

import gdefalll.Grammars.BBGrammarLexer;
import gdefalll.Grammars.BBGrammarParser;
import gdefalll.events.OkCancelGenerator;

import javax.swing.JTextPane;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.StyledDocument;
import java.awt.Color;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import org.antlr.runtime.*;

/**
 * Component to highlight ALLL BB tuples.
 * It checks syntax of written text.
 * License: BUT OPEN SOURCE LICENCE v.1
 * 
 * @author Jan Horacek <ihoracek@fit.vutbr.cz>
 */
public class ALLLBBHighlighter extends JTextPane implements KeyListener{
	/**
	 * Available modes of this component. UNIFIABLE mode accepts also '_' unificator (SIMPLE mode doesn't).
	 */
	public enum Mode {SIMPLE, UNIFIABLE};
	private static final long serialVersionUID = 1L;
	private boolean is_valid = false;
	private Mode mode = Mode.SIMPLE;
	/**
	 * Reaction on RETURN and ESC key-press. Listeners should register here.
	 */
	public OkCancelGenerator okc = new OkCancelGenerator();

	public ALLLBBHighlighter(Mode mode)
	{
		super();
		this.mode = mode;
		StyledDocument doc = getStyledDocument();
		addStylesToDocument(doc);
		addKeyListener(this);
	}
	
	protected void addStylesToDocument(StyledDocument doc) {
        //Initialize some styles.
        Style def = StyleContext.getDefaultStyleContext().
                        getStyle(StyleContext.DEFAULT_STYLE);

        Style s = doc.addStyle("token" + BBGrammarLexer.LEFT_PAREN, def);
        StyleConstants.setFontFamily(s, "SansSerif");
        StyleConstants.setBold(s, true);
        StyleConstants.setItalic(s, false);
        StyleConstants.setStrikeThrough(s, false);
        StyleConstants.setForeground(s, new Color(0, 69, 134));
        
        s = doc.addStyle("token" + BBGrammarLexer.RIGHT_PAREN, def);
        StyleConstants.setFontFamily(s, "SansSerif");
        StyleConstants.setBold(s, true);
        StyleConstants.setItalic(s, false);
        StyleConstants.setStrikeThrough(s, false);
        StyleConstants.setForeground(s, new Color(0, 69, 134));
        
        s = doc.addStyle("token" + BBGrammarLexer.NUMBER, def);
        StyleConstants.setFontFamily(s, "SansSerif");
        StyleConstants.setBold(s, false);
        StyleConstants.setItalic(s, false);
        StyleConstants.setStrikeThrough(s, false);
        StyleConstants.setForeground(s, new Color(0, 0, 128));
        
        s = doc.addStyle("token" + BBGrammarLexer.WORD, def);
        StyleConstants.setFontFamily(s, "SansSerif");
        StyleConstants.setItalic(s, true);
        StyleConstants.setBold(s, false);
        StyleConstants.setStrikeThrough(s, false);
        StyleConstants.setForeground(s, new Color(128, 0, 0));
        
        s = doc.addStyle("token" + BBGrammarLexer.COMMA, def);
        StyleConstants.setFontFamily(s, "SansSerif");
        StyleConstants.setBold(s, false);
        StyleConstants.setItalic(s, false);
        StyleConstants.setStrikeThrough(s, false);
        StyleConstants.setForeground(s, new Color(0, 0, 0));
        
        s = doc.addStyle("token" + BBGrammarLexer.REGISTER, def);
        StyleConstants.setFontFamily(s, "SansSerif");
        StyleConstants.setBold(s, true);
        StyleConstants.setItalic(s, true);
        StyleConstants.setStrikeThrough(s, false);
        StyleConstants.setForeground(s, new Color(153, 153, 255));
        
        s = doc.addStyle("token" + BBGrammarLexer.UNIFICATOR, def);
        StyleConstants.setFontFamily(s, "SansSerif");
        StyleConstants.setBold(s, true);
        StyleConstants.setItalic(s, false);
        StyleConstants.setStrikeThrough(s, false);
        StyleConstants.setForeground(s, new Color(153, 153, 255));
        
        s = doc.addStyle("token" + BBGrammarLexer.WS, def);
        StyleConstants.setFontFamily(s, "SansSerif");
        StyleConstants.setBold(s, false);
        StyleConstants.setItalic(s, false);
        StyleConstants.setStrikeThrough(s, false);
        StyleConstants.setForeground(s, new Color(0, 0, 0));
        
        s = doc.addStyle("problem", def);
        StyleConstants.setFontFamily(s, "SansSerif");
        StyleConstants.setBold(s, true);
        StyleConstants.setItalic(s, false);
        StyleConstants.setStrikeThrough(s, true);
        StyleConstants.setForeground(s, new Color(224, 0, 224));
    }
	
	/**
	 * @return mode of this component
	 * @see ALLLBBHighlighter.Mode
	 */
	private Mode getMode()
	{
		return mode;
	}
	
	@Override
	public void setText(String text)
	{
		super.setText(text);
		formatText();
	}
	
	/**
	 * Highlight syntax of this text.
	 */
	synchronized private void formatText()
	{
		is_valid = false;
		String text = getText();
		StyledDocument doc = this.getStyledDocument();
		ANTLRStringStream in = new ANTLRStringStream(text);
		BBGrammarLexer lexer = new BBGrammarLexer(in);
		Token token;
//		System.out.println(text);
		int last_position = 0;
		while (true) {
			token = lexer.nextToken();
			if (token.getType() == Token.EOF) {
				break;
			}
			if (token.getType() == BBGrammarLexer.UNIFICATOR && getMode() != Mode.UNIFIABLE) {
				doc.setCharacterAttributes(last_position, text.length() - last_position, doc.getStyle("problem"), false);
				break;
			}
			if (last_position < token.getCharPositionInLine())
			{
				doc.setCharacterAttributes(last_position, text.length() - last_position, doc.getStyle("problem"), false);
				return;
			}
			last_position = token.getCharPositionInLine() + token.getText().length();
			doc.setCharacterAttributes(token.getCharPositionInLine(), token.getText().length(), doc.getStyle("token" + token.getType()), false);
		}
		if (text.length() > last_position)
		{
			doc.setCharacterAttributes(last_position, text.length() - last_position, doc.getStyle("problem"), false);
			return;
		}
		in = new ANTLRStringStream(text);
		lexer = new BBGrammarLexer(in);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		BBGrammarParser parser = new BBGrammarParser(tokens);
		try {
			parser.parse();
		} catch (RecognitionException e) {
			System.out.println("Problem at:" + e.charPositionInLine);
		}
//		System.out.println("Errors:" + parser.getNumberOfSyntaxErrors());
		if (parser.getNumberOfSyntaxErrors() > 0)
		{
			int error_position = parser.getErrorStartingPosition();
			doc.setCharacterAttributes(error_position, text.length() - error_position, doc.getStyle("problem"), false);
			return;
		}
		is_valid = true;
	}
	
	@Override
	public void keyTyped(KeyEvent e) {
		if (e.getKeyCode() == KeyEvent.VK_ENTER)
		{
			e.consume();
		}
	}

	@Override
	public void keyPressed(KeyEvent e) {
		if (e.getKeyCode() == KeyEvent.VK_ENTER)
		{
			e.consume();
		}
	}

	@Override
	public void keyReleased(KeyEvent e) {
		if (e.getKeyCode() == KeyEvent.VK_ENTER)
		{
			e.consume();
			okc.fireOk();
		}
		else if (e.getKeyCode() == KeyEvent.VK_ESCAPE)
		{
			okc.fireCancel();
		}
		else
		{
			formatText();
		}
	}
	
	/* (non-Javadoc)
	 * return if the text inside this component is valid
	 * @see java.awt.Component#isValid()
	 */
	public boolean isValid()
	{
		return is_valid;
	}
}
